from inspect import getmro

from marshmallow.schema import Schema, SchemaMeta, SchemaOpts

from .converter import BaseConverter
from .registry import registry
from typing import Dict, Any


class AnnotationSchemaOpts(SchemaOpts):
    """
    marshmallow-annotations specific SchemaOpts implementation, provides:

    - converter_factory
    - registry
    - register_as_scheme
    - target
    - field_configs
    - converter
    """

    def __init__(self, meta, schema=None, ordered=False):
        super().__init__(meta)
        self.__sentinel = object()
        self.field_configs: Dict[str, Dict[str, Any]] = {}

        self._process(meta, schema)
        self._finalize()
        self.converter = self.converter_factory(registry=self.registry)

        if schema is not None and self.register_as_scheme and hasattr(self, "target"):
            self.converter.registry.register_scheme_factory(self.target, schema)

        del self.__sentinel

    def _process(self, meta, schema):
        self._extract_from_parents(schema, self._extract_from)
        self._extract_from(meta)
        self._gather_field_configs(schema, meta)

    def _extract_from_parents(self, schema, f):
        # can't look at just next parent
        # they may or may be an AnnotationSchema
        # so we need to walk backwards from object to get the correct
        # combination of settings
        for parent in reversed(getmro(schema)):
            opts = getattr(parent, "opts", self.__sentinel)
            if opts is self.__sentinel:
                continue

            f(opts)

    def _extract_from(self, source):
        if hasattr(source, "converter_factory"):
            self.converter_factory = source.converter_factory
        if hasattr(source, "register_as_scheme"):
            self.register_as_scheme = source.register_as_scheme
        if hasattr(source, "target"):
            self.target = source.target
        if hasattr(source, "registry"):
            self.registry = source.registry

    def _gather_field_configs(self, schema, meta):
        def merge_field_configs(opts):
            field_configs = getattr(opts, "field_configs", self.__sentinel)

            if field_configs is self.__sentinel:
                return

            for k, v in field_configs.items():
                config = self.field_configs.setdefault(k, {})
                config.update(dict(v))

        self._extract_from_parents(schema, merge_field_configs)

        defaults = getattr(meta, "Fields", self.__sentinel)
        if defaults is self.__sentinel:
            return

        for k, v in defaults.__dict__.items():
            if k.startswith("_"):
                continue

            self.field_configs.setdefault(k, {}).update(v)

    def _finalize(self):
        self.converter_factory = getattr(self, "converter_factory", BaseConverter)
        self.register_as_scheme = getattr(self, "register_as_scheme", False)
        self.registry = getattr(self, "registry", registry)


class AnnotationSchemaMeta(SchemaMeta):
    @classmethod
    def get_declared_fields(mcls, klass, cls_fields, inherited_fields, dict_cls):
        fields = super().get_declared_fields(
            klass, cls_fields, inherited_fields, dict_cls
        )

        target = getattr(klass.opts, "target", None)

        if target is None:
            return fields

        converter = klass.opts.converter

        # ignore anything explicitly declared on this scheme
        # or any parent scheme, also ignore anything explicitly
        # passed into exclude
        ignore = set(fields) | set(klass.opts.exclude)
        fields.update(converter.convert_all(target, ignore, klass.opts.field_configs))

        return fields


class AnnotationSchema(Schema, metaclass=AnnotationSchemaMeta):
    """
    Base class for creating annotation schema with::

        from marshmallow_annotations import AnnotationSchema
        from my.app.entities import Artist

        class ArtistScheme(AnnotationSchema):
            class Meta:
                target = Artist
                register_as_scheme = True

    """

    OPTIONS_CLASS_TYPE = AnnotationSchemaOpts

    @classmethod
    def OPTIONS_CLASS(cls, meta, **kwargs):
        return cls.OPTIONS_CLASS_TYPE(meta, cls, **kwargs)



import sys

IS_PY36 = sys.version_info[:2] == (3, 6)

if IS_PY36:

    def _is_class_var(typehint):
        from typing import _ClassVar  # type: ignore

        try:
            return isinstance(typehint, _ClassVar)
        except TypeError:  # pragma: no branch
            return False

    def _get_base(typehint):
        return typehint.__base__


else:

    def _is_class_var(typehint):
        # follow the lead of stdlib's dataclass implementation
        from typing import ClassVar, _GenericAlias  # type: ignore

        return typehint is ClassVar or (
            type(typehint) is _GenericAlias and typehint.__origin__ is ClassVar
        )

    def _get_base(typehint):
        return typehint.__origin__



from enum import Enum, EnumMeta

from marshmallow import ValidationError
from marshmallow.fields import Field
from .marshmallow_annotations import registry


BASE_TYPES = [int, float, str, tuple]


class EnumField(Field):
    default_error_messages = {
        'invalid_value': 'Invalid enum value for {cls}: {inpt}',
        'invalid_type': 'Enum type is invalid: {inpt} is not {enum_type}'
    }

    def __init__(self, cls, *args,
                 dump_by_value=True, load_by_value=True, **kwargs):
        self._enum_cls = cls
        self._enum_type = [_type
                           for _type in BASE_TYPES
                           if issubclass(cls, _type)][0]
        self.dump_by_value = dump_by_value
        self.load_by_value = load_by_value
        kwargs['enum'] = (list(cls._value2member_map_.keys())
                          if load_by_value else cls._member_names_)
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj):
        assert isinstance(value, self._enum_cls)
        return value.value if self.dump_by_value else value.name

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        if not self.load_by_value:
            try:
                return self._enum_cls[value]
            except KeyError:
                self.fail('invalid_value', inpt=value)
            except TypeError:
                self.fail('invalid_type', inpt=value)
        try:
            value = self._enum_type(value)
        except (TypeError, ValueError):
            self.fail('invalid_type', inpt=value)
        try:
            return self._enum_cls._value2member_map_[value]
        except KeyError:
            pass
        except TypeError:
            for k, v in self._enum_cls._value2member_map_.items():
                if v == value:
                    return self._enum_cls[k]
        self.fail('invalid_value', inpt=value)

    def fail(self, key, **kwargs):
        if key in self.default_error_messages:
            msg = self.default_error_messages[key].format(
                inpt=kwargs['inpt'],
                enum_type=self._enum_type.__name__,
                cls=self._enum_cls.__name__)
            raise ValidationError(msg)
        super().fail(key, **kwargs)


class RegisteredEnumMeta(EnumMeta):
    def __new__(metacls, name, bases, cls_dict):
        by_value = cls_dict.pop('__by_value__', True)
        dump_by_value = cls_dict.pop('__dump_by_value__', by_value)
        load_by_value = cls_dict.pop('__load_by_value__', by_value)
        if (not any(base_type in base_cls.mro()
                    for base_cls in bases
                    for base_type in BASE_TYPES)
                and name != 'RegisteredEnum'):
            bases = (str, *bases)
        cls = super().__new__(metacls, name, bases, cls_dict)
        cls.load_by_value = load_by_value
        cls.dump_by_value = dump_by_value
        return cls


class RegisteredEnum(Enum, metaclass=RegisteredEnumMeta):
    def __init_subclass__(cls):
        def _enum_field_converter(converter, subtypes, opts):
            return EnumField(cls,
                             load_by_value=cls.load_by_value,
                             dump_by_value=cls.dump_by_value)

        registry.register(cls, _enum_field_converter)



from typing import AbstractSet, Iterable

from attr import NOTHING, Attribute, Factory

from marshmallow import missing, post_load

from ..base import GeneratedFields, NamedConfigs
from ..exceptions import AnnotationConversionError
from ..scheme import AnnotationSchema, BaseConverter

__all__ = ("AttrsConverter", "AttrsSchema")

__SENTINEL = object()


def _is_attrs(target):
    return getattr(target, "__attrs_attrs__", __SENTINEL) is not __SENTINEL


def _get_attr_from_attrs(attrs: Iterable[Attribute], name: str) -> Attribute:
    attrs = [a for a in attrs if a.name == name]
    return attrs[0]


def _should_include_default(attr):
    if not attr.init:
        return False

    # an attrs factory cannot be copied over to a marshmallow field despite
    # marshmallow fields being able to accept a callable there because some
    # attrs factories accept the newly created instance as an argument
    # which is something marshmallow doesn't support

    # see following issue for mypy ignore
    # https://github.com/python/mypy/issues/3060
    return attr.default != NOTHING and not isinstance(  # type: ignore
        attr.default, Factory
    )


class AttrsConverter(BaseConverter):
    def convert_all(
        self,
        target: type,
        ignore: AbstractSet[str] = frozenset([]),  # noqa
        configs: NamedConfigs = None,
    ) -> GeneratedFields:
        self._ensure_all_hints_are_attribs(target, ignore)
        return super().convert_all(target, ignore, configs)

    def _get_field_defaults(self, target):
        return {
            a.name: a.default
            for a in target.__attrs_attrs__
            if _should_include_default(a)
        }

    def _preprocess_typehint(self, typehint, kwargs, field_name, target):
        # while this seems contradictory to this converter, we need to
        # ignore attrs specific actions when a container type, e.g. a List[T],
        # contains a non-attrs manufactured type because this is recursively
        # called during schema generation.
        # however those types will still require an entry in the registry
        # is handled outside this converter
        if not _is_attrs(target):
            return

        attr = _get_attr_from_attrs(target.__attrs_attrs__, field_name)

        if attr.default != NOTHING:
            # default to optional even if the typehint isn't
            # but don't override the user saying otherwise
            kwargs.setdefault("required", False)
            kwargs.setdefault("missing", missing)

    def _postprocess_typehint(self, typehint, kwargs, field_name, target):
        # see _preprocess_typehint for details
        if not _is_attrs(target):
            return

        attr = _get_attr_from_attrs(target.__attrs_attrs__, field_name)

        if not attr.init:
            # force into dump only mode if the field won't be accepted
            # into __init__ -- even if the user told us otherwise
            kwargs["dump_only"] = True

        if attr.metadata:
            kwargs.update(attr.metadata)

    def _ensure_all_hints_are_attribs(self, target, ignore):
        # This would happen if an attrs handled class was subclassed by
        # a plain ol' python class and added more non-ignored type hinted
        # fields. In theory we could handle this but we won't
        hints = {k for k, _ in self._get_type_hints(target, ignore)}
        attribs = {a.name for a in target.__attrs_attrs__ if a.name not in ignore}

        if hints != attribs:
            raise AnnotationConversionError(
                f"Mismatched hints and attr.ibs. Got hints: {hints} "
                f"but attr.ibs {attribs}"
            )


class AttrsSchema(AnnotationSchema):
    """
    Schema for handling ``attrs`` based targets, adds automatic load conversion
    into the target class and specifies the
    :class:`~marshmallow_annotations.ext.attrs.AttrsConverter` as the converter
    factory.
    """

    class Meta:
        converter_factory = AttrsConverter

    @post_load
    def make_object(self, data, many=False, partial=False):
        return self.opts.target(**data)



from .converter import BaseConverter
from .exceptions import AnnotationConversionError, MarshmallowAnnotationError
from .registry import TypeRegistry, field_factory, registry, scheme_factory
from .scheme import AnnotationSchema, AnnotationSchemaMeta

__version__ = "2.4.0.post1"
__author__ = "Alec Nikolas Reiter"
__license__ = "MIT"



from typing import AbstractSet, Union, get_type_hints

import marshmallow

from ._compat import _get_base, _is_class_var
from .base import (
    AbstractConverter,
    ConfigOptions,
    GeneratedFields,
    NamedConfigs,
    TypeRegistry,
)
from .registry import registry

NoneType = type(None)


def _is_optional(typehint):
    # only supports single type optionals/unions
    # as for the implementation... look, don't ask me
    return (
        hasattr(typehint, "__origin__")
        and typehint.__origin__ is Union
        and len(typehint.__args__) == 2
        and NoneType in typehint.__args__  # type: ignore
    )


def _extract_optional(typehint):
    """Given Optional[X] return X."""
    optional_types = [t for t in typehint.__args__ if t is not NoneType]  # type: ignore
    return optional_types[0]


def should_include(typehint):
    return not _is_class_var(typehint)


class BaseConverter(AbstractConverter):
    """
    Default implementation of :class:`~marshmallow_annotations.base.AbstractConverter`.

    Handles parsing types for type hints and mapping those type hints into
    marshmallow field instances by way of a
    :class:`~marshmallow_annotations.base.TypeRegistry` instance.

    :versionchanged: 2.1.0 Added non-public hook ``_get_field_defaults``

    :versionchanged: 2.2.0 Added non-public hooks ``_preprocess_typehint``
        and ``_postprocess_typehint``
    """

    def __init__(self, *, registry: TypeRegistry = registry) -> None:
        self.registry = registry

    def convert(
        self,
        typehint: type,
        opts: ConfigOptions = None,
        *,
        field_name: str = None,
        target: type = None
    ) -> marshmallow.fields.FieldABC:
        opts = opts if opts is not None else {}
        return self._field_from_typehint(
            typehint, opts, field_name=field_name, target=target
        )

    def convert_all(
        self,
        target: type,
        ignore: AbstractSet[str] = frozenset([]),  # noqa
        configs: NamedConfigs = None,
    ) -> GeneratedFields:
        configs = configs if configs is not None else {}
        for k, default in self._get_field_defaults(target).items():
            configs[k] = {"missing": default, **configs.get(k, {})}
        return {
            k: self.convert(v, configs.get(k, {}), field_name=k, target=target)
            for k, v in self._get_type_hints(target, ignore)
        }

    def is_scheme(self, typehint: type) -> bool:
        constructor = self.registry.get(typehint)
        return getattr(constructor, "__is_scheme__", False)

    def _field_from_typehint(
        self, typehint, kwargs=None, *, field_name: str = None, target: type = None
    ):
        # need that immutable dict in the stdlib pls
        kwargs = kwargs if kwargs is not None else {}
        self._preprocess_typehint(typehint, kwargs, field_name, target)

        # sane defaults
        allow_none = False
        required = True
        missing = marshmallow.missing

        if _is_optional(typehint):
            allow_none = True
            required = False
            missing = None
            typehint = _extract_optional(typehint)

        # set this after optional check
        subtypes = getattr(typehint, "__args__", ())

        if subtypes != ():
            typehint = _get_base(typehint)

        kwargs.setdefault("allow_none", allow_none)
        kwargs.setdefault("required", required)
        kwargs.setdefault("missing", missing)

        self._postprocess_typehint(typehint, kwargs, field_name, target)
        field_constructor = self.registry.get(typehint)
        return field_constructor(self, subtypes, kwargs)

    def _get_type_hints(self, item, ignore):
        """
        Helper to gather typehints from entire MRO.

        :versionchanged: 2.2.0 Push filtering of typehints into this method,
            return type is now Iterable[Tuple[str, type]]
        """
        hints = {}
        for parent in item.__mro__[::-1]:
            hints.update(get_type_hints(parent))
        return [
            (k, v) for (k, v) in hints.items() if k not in ignore and should_include(v)
        ]

    def _get_field_defaults(self, item):
        """
        Non-public hookpoint to read default values for all fields from the target
        """
        return {}

    def _preprocess_typehint(self, typehint, kwargs, field_name, target):
        """
        Non-public hookpoint for any preprocessing of typehint parsing
        for a given field
        """
        pass

    def _postprocess_typehint(self, typehint, kwargs, field_name, target):
        """
        Non-public hookpoint for any postprocessing of typehint parsing
        for a given field.
        """
        pass



from collections import OrderedDict

import attr
from sqlalchemy import inspect
from sqlalchemy.ext.declarative import declared_attr

from .marsh_schema import attr_with_schema


def model_with_schema(maybe_cls=None, **kwargs):
    """Make the sqlalchemy class an attr.s
    Also, updates annotations for `declared_attr`s
    """
    def decorator(cls):
        these = OrderedDict()
        these_with_defaults = {}
        if not hasattr(cls, "__annotations__"):
            cls.__annotations__ = {}
        for name, column in inspect(cls).columns.items():
            if column.nullable:
                these_with_defaults[name] = attr.ib(default=None)
            else:
                these[name] = attr.ib()
            if name in cls.__annotations__:
                continue

            try:
                tp = column.type.python_type
            except NotImplementedError:
                pass
            else:
                cls.__annotations__[name] = tp
                continue

            for base in cls.__bases__:
                if name not in base.__dict__:
                    continue
                if not isinstance(base.__dict__[name], declared_attr):
                    continue
                cls.__annotations__[name] = base.__dict__[
                    name].fget.__annotations__['return']

        these.update(these_with_defaults)

        return attr_with_schema(
            **{"register_as_scheme": True, "strict": True, **kwargs})(
                attr.s(cls, these=these, init=False))

    if maybe_cls is None:
        return decorator
    else:
        return decorator(maybe_cls)



from datetime import date, datetime, time, timedelta
from decimal import Decimal
from typing import Dict, List, Tuple, Union
from uuid import UUID

from marshmallow import fields
from marshmallow.base import FieldABC, SchemaABC

from ._compat import _get_base
from .base import AbstractConverter, ConfigOptions, FieldFactory, TypeRegistry
from .exceptions import AnnotationConversionError


def _is_generic(typehint: type) -> bool:
    # this *could* be isinstance(typehint, (Generic, _GenericAlias)) but
    # this works out better given that __origin__ isn't likely to go away
    # the way _GenericAlias might
    return getattr(typehint, "__origin__", None) is not None


def field_factory(field: FieldABC) -> FieldFactory:
    """
    Maps a marshmallow field into a field factory
    """

    def _(
        converter: AbstractConverter, subtypes: Tuple[type], opts: ConfigOptions
    ) -> FieldABC:
        return field(**opts)

    _.__name__ = f"{field.__name__}FieldFactory"
    return _


def scheme_factory(scheme_name: str) -> FieldFactory:
    """
    Maps a scheme or scheme name into a field factory
    """

    def _(
        converter: AbstractConverter, subtypes: Tuple[type], opts: ConfigOptions
    ) -> FieldABC:
        return fields.Nested(scheme_name, **opts)

    _.__name__ = f"{scheme_name}FieldFactory"
    _.__is_scheme__ = True  # type: ignore
    return _


def _list_converter(
    converter: AbstractConverter, subtypes: Tuple[type], opts: ConfigOptions
) -> FieldABC:
    if converter.is_scheme(subtypes[0]):
        opts["many"] = True
        return converter.convert(subtypes[0], opts)
    sub_opts = opts.pop("_interior", {})
    return fields.List(converter.convert(subtypes[0], sub_opts), **opts)


class DefaultTypeRegistry(TypeRegistry):
    """
    Default implementation of :class:`~marshmallow_annotations.base.TypeRegistry`.

    Provides default mappings of:

    - bool -> fields.Boolean
    - date -> fields.Date
    - datetime -> fields.DateTime
    - Decimal -> fields.Decimal
    - float -> fields.Float
    - int -> fields.Integer
    - str -> fields.String
    - time -> fields.Time
    - timedelta -> fields.TimeDelta
    - UUID -> fields.UUID
    - dict -> fields.Dict
    - typing.Dict -> fields.Dict

    As well as a special factory for typing.List[T] that will generate either
    fields.List or fields.Nested
    """

    _registry = {
        k: field_factory(v)
        for k, v in {
            bool: fields.Boolean,
            date: fields.Date,
            datetime: fields.DateTime,
            Decimal: fields.Decimal,
            float: fields.Float,
            int: fields.Integer,
            str: fields.String,
            time: fields.Time,
            timedelta: fields.TimeDelta,
            UUID: fields.UUID,
            dict: fields.Dict,
            Dict: fields.Dict,
        }.items()
    }

    # py36, py37 compatibility, register both out of praticality
    _registry[List] = _list_converter
    _registry[list] = _list_converter

    def __init__(self, registry: Dict[type, FieldFactory] = None) -> None:
        if registry is None:
            registry = {}

        self._registry = {**self._registry, **registry}

    def register(self, target: type, constructor: FieldFactory) -> None:
        self._registry[target] = constructor

    def get(self, target: type) -> FieldFactory:
        converter = self._registry.get(target)
        if converter is None and _is_generic(target):
            converter = self._registry.get(_get_base(target))

        if converter is None:
            raise AnnotationConversionError(f"No field factory found for {target!r}")
        return converter

    def register_field_for_type(self, target: type, field: FieldABC) -> None:
        self.register(target, field_factory(field))

    def register_scheme_factory(
        self, target: type, scheme_or_name: Union[str, SchemaABC]
    ) -> None:
        self.register(target, scheme_factory(scheme_or_name))

    def has(self, target: type) -> bool:
        return target in self._registry


registry = DefaultTypeRegistry()



import attr
from copy import copy
from functools import wraps

from .marshmallow_annotations.ext.attrs import AttrsSchema


def attr_with_schema(**kwargs):
    def decorator(cls):
        bases = []
        if hasattr(cls, 'Schema'):
            bases.append(cls.Schema)
        bases.append(AttrsSchema)

        schema_meta = getattr(cls, 'SchemaMeta', object)
        fields = attr.fields(cls)

        class Schema(*bases):
            class Meta(schema_meta):
                locals().update(kwargs)
                target = cls

        for field in fields:
            if field.default == attr.NOTHING:
                Schema._declared_fields[field.name].required = True

        cls.schema = Schema
        Schema.__name__ = cls.__name__ + "Schema"

        def attr_iter(self):
            return iter(self.__dict__.items())

        cls.__iter__ = attr_iter
        return cls
    return decorator


def derive(from_,
           *,
           exclude=None,
           derive_schema=True,
           derive_schema_meta=True):
    if exclude is None:
        exclude = []
    exclude = set(exclude)

    def decorator(cls):
        @wraps(cls, updated=[])
        class Wrapped(cls):
            locals().update({attr.name: attr.default
                             for attr in from_.__attrs_attrs__
                             if attr.name not in exclude})
        Wrapped.__annotations__ = copy(getattr(cls, '__annotations__', {}))
        for k in Wrapped.__annotations__.keys():
            if hasattr(cls, k):
                setattr(Wrapped, k, getattr(cls, k))
        if derive_schema and hasattr(from_, 'Schema'):
            if hasattr(cls, 'Schema'):
                class Schema(cls.Schema, from_.Schema):
                    pass
            else:
                Schema = from_.Schema
            Wrapped.Schema = Schema

        if derive_schema_meta and hasattr(from_, 'SchemaMeta'):
            if hasattr(cls, 'SchemaMeta'):
                class SchemaMeta(cls.SchemaMeta, from_.SchemaMeta):
                    pass
            else:
                SchemaMeta = from_.SchemaMeta
            Wrapped.SchemaMeta = SchemaMeta

        Wrapped.__annotations__.update({attr.name: attr.type
                                        for attr in from_.__attrs_attrs__
                                        if attr.name not in exclude})
        return Wrapped
    return decorator



"""Specialized components for Python 3.6 NamedTuple annotations."""

import marshmallow

from marshmallow_annotations.scheme import (
    AnnotationSchema,
    AnnotationSchemaOpts,
    BaseConverter,
)


__all__ = ("NamedTupleConverter", "NamedTupleSchemaOpts", "NamedTupleSchema")


class NamedTupleConverter(BaseConverter):
    def _get_field_defaults(self, item):
        return getattr(item, "_field_defaults", {})


class NamedTupleSchemaOpts(AnnotationSchemaOpts):
    """
    NamedTuple specific AnnotationSchemaOpts, additionally provides:

    - dump_default_fields
    """

    def __init__(self, meta, *args, **kwargs):
        super().__init__(meta, *args, **kwargs)
        self.dump_default_fields = getattr(meta, "dump_default_fields", True)


class NamedTupleSchema(AnnotationSchema):
    """
    Derived class for creating typing.NamedTuple schema with automatic
    post-load conversion to namedtuple instances.
    """

    OPTIONS_CLASS_TYPE = NamedTupleSchemaOpts

    class Meta:
        converter_factory = NamedTupleConverter

    @marshmallow.post_load
    def make_namedtuple(self, data):
        """Post load, deserialize to target namedtuple class."""
        return self.opts.target(**data)

    @marshmallow.post_dump
    def remove_optional(self, data):
        """Post dump, strip default fields from serialized output."""
        if self.opts.dump_default_fields:
            return data
        else:
            default_values = self.opts.converter._get_field_defaults(self.opts.target)
            return {k: v for k, v in data.items() if v != default_values.get(k)}



from abc import ABC, abstractmethod
from typing import AbstractSet, Any, Callable, Dict, Optional, Tuple, Union

from marshmallow.base import FieldABC, SchemaABC

ConfigOptions = Optional[Dict[str, Any]]
NamedConfigs = Optional[Dict[str, ConfigOptions]]
GeneratedFields = Dict[str, FieldABC]


class AbstractConverter(ABC):
    """
    Converters handle gathering type hints and consulting with a
    :class:`~marshmallow_annotations.base.TypeRegistry` in order to produce
    marshmallow field instances::


        from marshmallow_annotations import BaseConverter

        converter = BaseConverter(registry)
        converter.convert(int, {"required": False})
        # marshmallow.fields.Integer(...)


    This abstract class is provided primarily for type hinting purposes
    but also allows implementations outside of the default implementation in
    this library.
    """

    @abstractmethod
    def convert(
        self,
        typehint: type,
        opts: ConfigOptions = None,
        *,
        field_name: str = None,
        target: type = None
    ) -> FieldABC:
        """
        Used to convert a type hint into a :class:`~marshmallow.base.FieldABC`
        instance.

        :versionchanged: 2.2.0 Added field_name and target optional keyword
            only arguments
        """
        pass

    @abstractmethod
    def convert_all(
        self,
        target: type,
        ignore: AbstractSet[str] = frozenset([]),  # noqa: B008
        configs: NamedConfigs = None,
    ) -> GeneratedFields:
        """
        Used to transform a type with annotations into a dictionary mapping
        the type's attribute names to :class:`~marshmallow.base.FieldABC`
        instances.
        """
        pass

    @abstractmethod
    def is_scheme(self, typehint: type) -> bool:
        """
        Used to check if the typehint passed if associated to a scheme or
        a regular field constructor.
        """
        pass


FieldFactory = Callable[[AbstractConverter, Tuple[type], Dict[str, Any]], FieldABC]


class TypeRegistry(ABC):
    """
    Abstraction representation of a registry mapping Python types to marshmallow
    field types.

    This abstract class is provided primarily for type hinting purposes but also
    allows implementations outside of the default implementation in this library.
    """

    @abstractmethod
    def register(self, target: type, constructor: FieldFactory) -> None:
        """
        Registers a raw field factory for the specified type::

            from marshmallow import fields

            def custom(converter, subtypes, opts):
                return fields.Raw(**opts)

            registry.register(bytes, custom)


        """
        pass

    def field_factory(self, target: type) -> Callable[[FieldFactory], FieldFactory]:
        """
        Decorator form of register::

            @register.field_factor(bytes)
            def custom(converter, subtypes, opts):
                return fields.Raw(**opts)

        Returns the original function so it can be used again if needed.
        """

        def field_factory(constructor: FieldFactory) -> FieldFactory:
            self.register(target, constructor)
            return constructor

        return field_factory

    @abstractmethod
    def get(self, target: type) -> FieldFactory:
        """
        Retrieves a field factory from the registry. If it doesn't exist,
        this may raise a
        :class:`~marshmallow_annotations.exception.AnnotationConversionError`::

            registry.get(str)  # string field factory
            registry.get(object)  # raises AnnotationConversionError

        """
        pass

    @abstractmethod
    def register_field_for_type(self, target: type, field: FieldABC) -> None:
        """
        Registers a raw marshmallow field to be associated with a type::

            from typing import NewType
            Email = NewType("Email", str)

            registry.register_field_for_type(Email, EmailField)

        """
        pass

    @abstractmethod
    def register_scheme_factory(
        self, target: type, scheme_or_name: Union[str, SchemaABC]
    ) -> None:
        """
        Registers an existing scheme or scheme name to be associated with a type::

            from myapp.schema import ArtistScheme
            from myapp.entities import Artist

            registry.register_scheme_factory(Artist, ArtistScheme)

        """
        pass

    @abstractmethod
    def has(self, target: type) -> bool:
        """
        Allows safely checking if a type has a companion field mapped already::

            registry.has(int)     # True
            registry.has(object)  # False

        May also be used with in::

            int in registry     # True
            object in registry  # False
        """
        pass

    def __contains__(self, target: type) -> bool:  # pragma: no cover
        return self.has(target)



from setuptools import setup

setup(
    name='marshmallow-helpers',
    version='0.4.4',
    packages=['marshmallow_helpers',
              'marshmallow_helpers.marshmallow_annotations',
              'marshmallow_helpers.marshmallow_annotations.ext'],
    author_email="zidder@hilearn.io",
    description="Ease marshmallow schema creation",
    classifiers=['Programming Language :: Python :: 3',
                 'Development Status :: 3 - Alpha'],
    install_requires=["attrs==20.3.0",
                      "marshmallow>=3.5.1",
                      "webargs>=5.5.3"],
    extras_require={
        "model": ["sqlalchemy>=1.3.20"]
    },
    dependency_links=["https://github.com/hilearn/marshmallow-annotations"]
)
